#!/usr/bin/env python2
from __future__ import print_function


"""systime-tracker.py: monitor system time"""
__version__   = "1.1"
__author__    = "ed <systimetrk.py@ocv.me>"
__credits__   = ["stackoverflow.com"]
__license__   = "MIT"
__copyright__ = 2016


import os
import sys
import time
import ctypes


##
## defined by /usr/include/linux/time.h

CLOCK_REALTIME = 0
CLOCK_MONOTONIC_RAW = 4

class timespec(ctypes.Structure):
	_fields_ = [
		('tv_sec',  ctypes.c_long),
		('tv_nsec', ctypes.c_long)
	]

librt = ctypes.CDLL('librt.so.1', use_errno=True)
clock_gettime = librt.clock_gettime
clock_gettime.argtypes = [
	ctypes.c_int,
	ctypes.POINTER(timespec)
]

def rt_gettime(clock_src):
	t = timespec()
	if clock_gettime(clock_src, ctypes.pointer(t)) != 0:
		errno_ = ctypes.get_errno()
		raise OSError(errno_, os.strerror(errno_))
	
	return t.tv_sec + t.tv_nsec * 1e-9

## end of linux includes
##


def main():
	msg = u'startup mono {}, real {}'.format(
		rt_gettime(CLOCK_MONOTONIC_RAW),
		rt_gettime(CLOCK_REALTIME))
	print(msg)
	
	logfile = None
	if len(sys.argv) > 1:
		logfile = sys.argv[1]
	
	if not logfile:
		print('note: optional argument 1 = logfile')
	else:
		with open(logfile, 'ab') as f:
			f.write((msg + u'\n').encode('utf-8'))
	
	prev_mono = None
	prev_real = None
	prev_accu = None
	
	while True:
		this_mono = None
		this_real = None
		this_accu = 9001
		
		# will use monotonic prev/this delta to find shift in realtime,
		# so first find the most accurate correlation between mono/real
		for n in range(5):
			mono0 = rt_gettime(CLOCK_MONOTONIC_RAW)
			real  = rt_gettime(CLOCK_REALTIME)
			mono1 = rt_gettime(CLOCK_MONOTONIC_RAW)
			accuracy = mono1 - mono0
			
			if accuracy < this_accu:
				this_mono = mono0 + (mono1-mono0)/2
				this_real = real
				this_accu = accuracy
			
			if accuracy < 0.00005:
				# usually first try with a cuo2duo
				break
		
		if prev_mono is not None:
			delta_mono = this_mono - prev_mono
			delta_real = this_real - prev_real
			clock_diff = delta_real - delta_mono
			max_diff = (prev_accu + this_accu) * 1.05
			
			if abs(clock_diff) > max_diff:
				msg = u'real skip {} sec from {} to {} (d {}), from mono {} to {} (d {}), accuracy {} + {} => {}'.format(
					*list(u'\033[36m{:.5f}\033[0m'.format(x) for x in [
						clock_diff,
						prev_real, this_real, delta_real,
						prev_mono, this_mono, delta_mono,
						prev_accu, this_accu, max_diff
					]))
				
				print(msg)
				if logfile:
					with open(logfile, 'ab') as f:
						f.write((msg + u'\n').encode('utf-8'))
		
		prev_mono = this_mono
		prev_real = this_real
		prev_accu = this_accu
		time.sleep(0.1)


if __name__ == '__main__':
	main()
